/**
 * DiscreteScale.java
 */
package UTEStatsViewer;
import GraphicPlotter.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
 * Y-axis scale of discrete values.
 */

public final class DiscreteScale extends AxisScale {
   private static final int tickWidth = 3;
   private static final int tickHeight = 1;

   /** The tick values. */
   private DiscreteScaleValues values;

   /** The user coordinate position of the lowest graphic element. */
   private int valueOffset;

   /** The user coordinate separation between discrete values. */
   private int valueSep;

   /** The user coordinate height of a discrete value. */
   private int valueHeight;


   /** Hide default constructor. */
   private DiscreteScale() {}


   /**
    * Construct a discrete scale.
    * @param values The values of the discrete scale.
    * @param valueOffset User coordinate position of the lowest
    *   graphic element.
    * @param valueHeight User coordinate separation between discrete
    *   values.
    * @param valueSep User coordinage height of a discrete value.
    */
   public DiscreteScale(DiscreteScaleValues values, int valueOffset,
         int valueHeight, int valueSep) {
      // save the parameter values
      this.values = values;
      this.valueOffset = valueOffset;
      this.valueHeight = valueHeight;
      this.valueSep = valueSep;
      setFont(new Font("Monospaced", Font.PLAIN, 12));
   }


   /** Return the DiscreteScaleValues object of this scale. */
   public DiscreteScaleValues getDiscreteScaleValues() {
      return values;
   }


   /** Convert from user coordinates to descrete value index. */
   private final int UC2Idx(double uc) {
      // the tick values are centered in the extent of user
      // coordinate values that map to a descrete value
      int ret;

      //ret = ((int)(round(uc) - valueOffset + 1)/(valueHeight + valueSep);
      ret = (int)(Math.round(uc / (valueHeight + valueSep)));

      // handle user coordinate out of range gracefully
      ret = Math.max(0, Math.min(values.size() - 1, ret));

      return ret;
   }


   /**
    * Convert discrete value index to user coordinate of bottom of
    * range.
    */
   private final double Idx2UC(int idx) {
      return ((valueHeight + valueSep) * idx);
   }



   // Define the ViewScale abstract methods.

   /** Implement createCoordinateAdjuster. */
   public CoordinateAdjuster createCoordinateAdjuster(double uc) {
      IncrementalChoice choice = new IncrementalChoice(
            values.getDiscreteScaleValueNamed(), UC2Idx(uc));
      return new CoordinateAdjuster(this, choice);
   }


   /** Implement extractLocation. */
   public double getAdjusterValue(CoordinateAdjuster la)
         throws RuntimeException {
      return Idx2UC(((IncrementalChoice)la.comp).getSelectedIndex());
   }

   /** Return desired insets. */
   public Insets getOutsets(int extent) {
      // assume that there is enough room
      return null;
   }



   // Override Component methods.

   public Dimension getMinimumSize() {
      return getPreferredSize();
   }


   public Dimension getPreferredSize() {
      // Need enough width to show the tick marks, the tick values,
      // and some space between them.
      FontMetrics fontMetrics = getFontMetrics(getFont());

      // set width to that of widest label
      String labels[] = values.getDiscreteScaleValues();
      int width = 0;
      for (int i = labels.length - 1; i >= 0; i--) {
         int labelWidth = fontMetrics.stringWidth(labels[i]);
         if (labelWidth > width) {
             width = labelWidth;
         }
      }

      // add width for 1) space between label and tick mark, 2) tick
      // mark, and 3) axis line
      width += fontMetrics.getMaxAdvance() + tickWidth + 1;

      return new Dimension(width, 1);
   }


   public void paint(Graphics g) {
      if (((int)extent) == 0) {
         return;   // scale extent not established
      }

      Dimension mySize = getSize();

      // determine size of font needed to fit label values between pixels
      double yIncr = (double) mySize.height / values.size();
      Font font = getFont();
      FontMetrics fm = getFontMetrics(font);
      if (yIncr < 4) {
         font = null;  // too little space between tick marks for labels
      } else {
         int fontHeight = fm.getMaxAscent() + fm.getMaxDescent() +
               fm.getLeading();
         if (fontHeight > yIncr){
            int fontSize = font.getSize();
            String fontName = font.getName();
            int fontStyle = font.getStyle();
            while ((fontSize > 1) && (fontHeight > yIncr)) {
               font = new Font(fontName, fontStyle, --fontSize);
               fm = getFontMetrics(font);
               fontHeight = fm.getMaxAscent() + fm.getMaxDescent() +
                     fm.getLeading();
            }
            if (fontSize < 1) {
               font = null;  // no font size small enough
            }
         }
      }

      // determine rightmost position of the labels and offset from tick
      // marks
      int xposRight = 0;
      int labelOffset = 0;
      if (font != null) {
         g.setFont(font);
         xposRight = mySize.width - 1 - tickWidth - fm.getMaxAdvance()/2;
         labelOffset = (fm.getAscent() - fm.getDescent() + 1) / 2;
      }

      // calculate Y-value to AWT coordinate parameters
      double pitch = (mySize.height - 1) / extent;
      double UCMax = origin + extent;

      // calculate range of task indicies to paint from clip values
      Rectangle clipBounds = g.getClipBounds();
      int discreteIdxLo = UC2Idx(UCMax -
            ((clipBounds.y + clipBounds.height - 1) / pitch)); 
      if (discreteIdxLo > 0) {
         // start at next label down in case the top most pixels of
         // its label are visible
         discreteIdxLo--;
      }
      int discreteIdxHi = UC2Idx(UCMax - (clipBounds.y / pitch)); 
      if (discreteIdxHi < (values.size() - 1)) {
         // start at next label up in case the bottom most pixels of
         // its label are visible
         discreteIdxHi++;
      }

      // draw vertical line for the axis
      g.drawLine(mySize.width - 1, clipBounds.y, mySize.width - 1,
            clipBounds.y + clipBounds.height - 1);

      // factors to determine user coordinate of the ith tick mark 
      int tickPosIncr = valueHeight + valueSep;

      // for each tick value visible in the clipping range
      for (int i = discreteIdxLo; i <= discreteIdxHi; i++) {
         int ypos;

         // draw the tick mark at top of extent of the discrete value
         ypos = (int)Math.round((UCMax - tickPosIncr * (i+1)) * pitch) -
               tickHeight / 2;
         g.fillRect(mySize.width - 1 - tickWidth, ypos, 
               tickWidth, tickHeight);

         // draw the label right justified against the tick mark with
         // the label centered in the extent of the discrete value
         if (font != null) {
            ypos = (int)Math.round((UCMax -
                tickPosIncr * (i + .5)) * pitch) + labelOffset;
            try {
               String label = values.getDiscreteScaleValue(i);
               g.drawString(label, xposRight - fm.stringWidth(label),
                     ypos);
            } catch (RuntimeException e) {
               // couldn't draw a label
            }
         }
      }
   }


   /**
    * Return the discrete index (0..DiscreteScaleValue.size())
    * to the given AWT location in this Component.  Returns -1 if the
    * location is not in the scale.
    */
   public int getIndexAt(int x, int y) {
      int idx = -1;
      if (contains(x, y)) {
         Dimension size = getSize();
         int yMax = size.height - 1;
         idx = (int)((yMax - y) * (long)values.size() / (yMax + 1));
      }
      return idx;
   }
}
