/**
 * MainFrame: the Frame class used by the program.
 */
package UTEStatsViewer;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.text.*;
import java.util.*;

public class MainFrame extends Frame implements ActionListener, MouseListener
{
   // group of windows this one is part of
   private WindowGroup windowGroup;

   // frame menu items
   private MenuItem openMenuItem;
   private MenuItem closeMenuItem;
   private MenuItem exitMenuItem;
   private MenuItem newWindowMenuItem;
   private Menu selectTableMenu = new Menu(Message.get(Message.selectTable));

   // container of the the table plots and its layout manager
   private Container cardPane;
   private CardLayout cardPaneLayout;

   // message area at bottom of frame
   private Label msgArea = new Label();

   // message area at top of frame
   private Label tableNameLabel = new Label();


   // table names are of the form "row name" + '/' + "col name"
   private static char tableNameSep = '/';


   // the object on which to invoke setRange when a range is selected
   private RangeCallback rangeCallback;

   /** Return the RangeCallback object being used. */
   public final RangeCallback getRangeCallback() {
      return rangeCallback;
   }

   /** Set the RangeCallback object to be used. */
   public final void setRangeCallback(RangeCallback rc) {
      rangeCallback = rc;
   }

   /** Set the RangeCallback object to be used to null. */
   public final void setRangeCallback() {
      rangeCallback = null;
   }


   // size of screen
   public static Dimension screenSize =
         Toolkit.getDefaultToolkit().getScreenSize();


   /** Create application frame that is not part of any group. */
   public MainFrame() {
      super();
      init(null);
   }


   /** Create application frame that is part of a group. */
   public MainFrame(WindowGroup windowGroup) {
      super();
      init(windowGroup);
   }


   /**
    * Create application frame with a title that is not part of
    * a group.
    */
   public MainFrame(String title) {
      this(title, null);
   }


   /** Create application frame with title that is part of a group. */
   public MainFrame(String title, WindowGroup windowGroup) {
      super(title);
      init(windowGroup);
   }


   /** Init method used by the various constructors. */
   private void init(WindowGroup windowGroup) {
      this.windowGroup = windowGroup;
      if (windowGroup != null) {
         addWindowListener(windowGroup);
      }
      setMenuBar(createMenuBar(windowGroup));
      tableNameLabel.setAlignment(Label.CENTER);
      add(tableNameLabel, BorderLayout.NORTH);
      add(msgArea, BorderLayout.SOUTH);

      // add container to which plotPanes are added
      cardPane = new ConcreteContainer();
      cardPaneLayout = new CardLayout();
      cardPane.setLayout(cardPaneLayout);
      add(cardPane, BorderLayout.CENTER);
   }


   /** Create and return a standard menu bar for this frame. */
   private MenuBar createMenuBar(WindowGroup windowGroup) {
      MenuBar menuBar = new MenuBar();

      Menu fileMenu = new Menu(Message.get(Message.file));
      menuBar.add(fileMenu);

      openMenuItem = new MenuItem(Message.get(Message.open));
      fileMenu.add(openMenuItem);
      openMenuItem.addActionListener(this);

      closeMenuItem = new MenuItem(Message.get(Message.close));
      fileMenu.add(closeMenuItem);
      closeMenuItem.addActionListener(this);

      if (windowGroup != null) {
         // Do not add "new window" and "exit" menu items if window
         // is not part of a group.
         fileMenu.addSeparator();

         newWindowMenuItem = new MenuItem(Message.get(Message.newWindow));
         fileMenu.add(newWindowMenuItem);
         newWindowMenuItem.addActionListener(this);

         fileMenu.addSeparator();

         exitMenuItem = new MenuItem(
            Message.get(windowGroup.getExitWhenEmptied() ? Message.exit :
                  Message.closeGroup)
         );
         fileMenu.add(exitMenuItem);
         exitMenuItem.addActionListener(this);
      }

      menuBar.add(selectTableMenu);

      return menuBar;
   }


   /** Load a frame with data */
   public void load(Reader reader) {
      TSVReaderTokenizer ts = new
            TSVReaderTokenizer(new BufferedReader(reader));
      StringBuffer sb;
      Vector colLabels;
      DiscreteScaleStrings rowLabels;
      HorzBarCategory categories[];
      int categoryCnt;
      HorzBarCategory unnamedCategory;
      long accumTime;
      String tableName = null;
      int rowCnt = 0, colCnt = 0;

      try {
         // clear existing frame
         selectTableMenu.removeAll();
         cardPane.removeAll();

         // read first table token
         sb = ts.nextToken();

         // for each table in the input stream
         while (sb != TSVReaderTokenizer.eofToken) {
            // first token is table name
            tableName = sb.toString().trim();
            msgArea.setText(tableName);
            String rowName = getRowName(tableName);
            String colName = getColName(tableName);
            sb = ts.nextToken();

            // add table to the container of them
            PlotPane plotPane = new PlotPane();
            cardPane.add(plotPane, tableName);
            if (cardPane.getComponentCount() == 1) {
               tableNameLabel.setText(tableName);
            }
            plotPane.addXScale(new TimeScale());

            // add table to select table menu
            {
               MenuItem menuItem = new MenuItem(tableName);
               selectTableMenu.add(menuItem);
               menuItem.addActionListener(this);
            }

            // rest of first row are column labels
            colLabels = new Vector();
            while ( (sb != TSVReaderTokenizer.eorToken) &&
                    (sb != TSVReaderTokenizer.eofToken) ) {
               colLabels.addElement(sb.toString().trim());
               sb = ts.nextToken();
            }
            if (sb == TSVReaderTokenizer.eorToken) {
               sb = ts.nextToken();  // consume eorToken
            }

            // create graphic category for each column
            categoryCnt = colLabels.size();
            categories = new HorzBarCategory[categoryCnt];
            for (int i = 0; i < categoryCnt; i++) {
               String colLabel = (String) colLabels.elementAt(i);
               categories[i] = new HorzBarCategory(
                     formatFullLabel(colName, colLabel), colLabel,
                     ColorPalette.getColor(i),  msgArea);
               plotPane.addLegend(categories[i]);
            }

            // prepare to read rows
            unnamedCategory = null;
            rowCnt = 0;
            rowLabels = new DiscreteScaleStrings();
            DiscreteScale discreteScale = new DiscreteScale(rowLabels,
                  HorzBar.OFFSET, HorzBar.HEIGHT, HorzBar.SEP);
            plotPane.addYScale(discreteScale);
            boolean isTimeBinScale = true;

            // for each row of a table
            while ( (sb != TSVReaderTokenizer.eofToken) &&
                    (sb != TSVReaderTokenizer.tableSepToken) ) {
               // first token of row is row label
               String rowLabel = sb.toString().trim();
               rowLabels.addElement(rowLabel);
               String rowFullLabel = formatFullLabel(rowName, rowLabel);
               isTimeBinScale = isTimeBinScale && isTimeBinValue(rowLabel);

               // for each column in a row
               sb = ts.nextToken();   // advance past row label
               colCnt = 0;
               accumTime = 0;
               while ( (sb != TSVReaderTokenizer.eofToken) &&
                       (sb != TSVReaderTokenizer.tableSepToken) &&
                       (sb != TSVReaderTokenizer.eorToken) ) {
                  // determine category
                  HorzBarCategory category;
                  if (colCnt < categories.length) {
                     category = categories[colCnt];
                  } else {
                     if (unnamedCategory == null) {
                        unnamedCategory = new HorzBarCategory("", "",
                              ColorPalette.getColor(categoryCnt),
                              msgArea);
                        plotPane.addLegend(unnamedCategory);
                     }
                     category = unnamedCategory;
                  }

                  // add graphic element to plot
                  long time = TimeValue.parseTime(sb);
                  if (time != 0) {
                     plotPane.addGraphic(category.createGraphic(
                           accumTime,
                           rowCnt * HorzBar.INCR + HorzBar.OFFSET, time,
                           rowFullLabel));
                     accumTime += time;
                  }

                  // advance to next column
                  colCnt++;
                  sb = ts.nextToken();
               }
               if (sb == TSVReaderTokenizer.eorToken) {
                  sb = ts.nextToken();  // consume eorToken
               }

               rowCnt++;
               if ((rowCnt % 1000) == 0) {
                  msgArea.setText(String.valueOf(rowCnt));
               }
            }

            if (sb == TSVReaderTokenizer.tableSepToken) {
               sb = ts.nextToken();  // consume tableSepToken
            }

            // if scale was a time bin scale, react to mouse clicks on it 
            if (isTimeBinScale) {
               discreteScale.addMouseListener(this);
            }

            msgArea.setText(Message.emptyString);
         }
      } catch (IOException e) {
         msgArea.setText(Message.get(e));
      } catch (NumberFormatException nfe) {
         String msgArgs[] = {tableName, String.valueOf(rowCnt),
               String.valueOf(colCnt), nfe.getMessage()};
         msgArea.setText(MessageFormat.format(Message.get(nfe), msgArgs));
      }

      validate();
   }


   /** Return whether a value is a time bin value. */
   private static boolean isTimeBinValue(String str) {
      int i = 0, size = str.length();
      boolean hadDecimalPoint = false;

      for (i = 0; i < size; i++) {
         char c = str.charAt(i);
         if (c == '.') {
            if (hadDecimalPoint) {
               return false;   // error: multiple decimal points
            } else {
               hadDecimalPoint = true;
            }
         } else if ( (c < '0') || (c > '9') ) {
            return false;   // error: character not a digit
         }
      }

      return hadDecimalPoint;
   }


   /** Return the column name component of a table name. */
   private static String getColName(String tableName) {
      int sepIdx = tableName.indexOf(tableNameSep); 
      return ( (sepIdx < 0) || (sepIdx >= (tableName.length()-1)) ) ?
            null : tableName.substring(sepIdx + 1).trim();
   }


   /** Return the row name component of a table name. */
   private static String getRowName(String tableName) {
      int sepIdx = tableName.indexOf(tableNameSep); 
      return (sepIdx < 1) ? null : tableName.substring(0, sepIdx).trim();
   }


   /** Format a fully identifying string for a row/column value. */
   private static String formatFullLabel(String labelName,
         String labelValue) {
      return (labelName == null) ? labelValue : labelName + "=" + labelValue;
   }


   /**
    * Set Component location to be cascaded from given location.
    */
   public static void cascadeLocation(Component c, int x, int y) {
      int newX = (x + 24) % (screenSize.width - 24);
      int newY = (y + 24) % (screenSize.height - 24);
      c.setLocation(newX, newY);
   }


  /** Implement ActionListener interface. */
  public void actionPerformed(ActionEvent e) {
     Object source = e.getSource();
     if (source == openMenuItem) {
        FileDialog fd = new
              FileDialog(this, Message.get(Message.open));
        fd.show();
        String chosenFile = fd.getFile();
        if ( (chosenFile == null) || (chosenFile.length() <= 0) ) {
           // file not chosen (FileDialog probably cancelled)
        } else {
           msgArea.setText(Message.emptyString);
           try {
              String fileName = fd.getDirectory() + fd.getFile();
              load(new FileReader(fileName));
              setTitle(fileName);
           } catch (FileNotFoundException fnfe) {
              msgArea.setText(Message.get(fnfe));
           }
        }
        fd.dispose();
     } else if (source == closeMenuItem) {
        setVisible(false);
        dispose();
     } else if (source == exitMenuItem) {
        windowGroup.disposeAll();
     } else if (source == newWindowMenuItem) {
        MainFrame frame = new MainFrame(windowGroup);
        Rectangle bounds = getBounds();
        frame.setSize(bounds.width, bounds.height);
        cascadeLocation(frame, bounds.x, bounds.y);
        frame.show();
     } else if (source instanceof MenuItem) {
        MenuItem menuItem = (MenuItem) source;
        if (menuItem.getParent() == selectTableMenu) {
           // table to view has been selected
           String tableName = e.getActionCommand();
           tableNameLabel.setText(tableName);
           cardPaneLayout.show(cardPane, tableName);
           msgArea.setText(Message.emptyString);
        }
     }
  }


  // implement MouseListener interface
  public void mouseClicked(MouseEvent e) {
     if (rangeCallback != null) {
        Component comp = e.getComponent();
        if ((e.getClickCount() == 2) && (comp instanceof DiscreteScale)) {
           // had double click on a scale, invoke the showRange() callback
           DiscreteScale ds = (DiscreteScale)comp;
           int idx = ds.getIndexAt(e.getX(), e.getY());
           if (idx >= 0) {
              // mouse location in scale
              DiscreteScaleValues dsv = ds.getDiscreteScaleValues();
              double start, end;

              // determine end of range as a double value
              {
                 String endStr = dsv.getDiscreteScaleValue(idx);
                 end = Double.valueOf(endStr).doubleValue();
              }

              // determine start of range as a double value
              if (idx == 0) {
                 start = 0;
              } else {
                 String startStr = dsv.getDiscreteScaleValue(idx - 1);
                 start = Double.valueOf(startStr).doubleValue();
              }

              // invoke the callback
              rangeCallback.showRange(start, end);
           }
        }
     }
  }


  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}
  public void mousePressed(MouseEvent e) {}
  public void mouseReleased(MouseEvent e) {
  }
}
